indexing

	description: "Root container for all child_definitions.%
	% This class helps to avoid special cases in DEFINED_ELEMENT_CHILDREN."


class

	CHILD_DEFINITION_ROOT


inherit

	CHILD_DEFINITION


creation {DEFINED_ELEMENT_CHILDREN}

	make


feature {NONE} -- Initialization

	make (a_child: CHILD_DEFINITION) is
		require
			child_not_void: a_child /= Void
		do
			child := a_child
		end


feature {DEFINED_ELEMENT} -- start tag transition

	is_allowed_child (state: ELEMENT_STATE;
							next_element: DEFINED_ELEMENT): BOOLEAN is
			-- We come here when our parent, a DEFINED_ELEMENT_CHILDREN,
			-- is asking if `next_element' is a valid child.
			-- Our parent expects us to keep track of the transition.
		require
			stack_is_valid: state.children /= Void
		local
			child_state: CHILD_STATE
		do
			state.set_authoritative_false (False)

			if state.children.is_empty then
				-- first time we're called, push ourselves upon the stack
				create child_state.make (Current)
				state.children.put (child_state)

				debug
					print ("CHILD_DEFINITION_ROOT.is_allowed_set%N")
				end
				Result := is_allowed_set (state, next_element)
			else
				debug
					print ("CHILD_DEFINITION_ROOT.is_allowed_stay%N")
				end
				Result := is_allowed_stay (state, next_element)
			end
		end

	is_allowed_end (state: ELEMENT_STATE): BOOLEAN is
			-- Is there no required child (anymore) in this child definition?
		do
			debug
				print ("CHILD_DEFINITION_ROOT.is_allowed_end%N")
			end

			if state.children.is_empty then
				-- child definition never activated, start scanning from start
				Result := not child.any_required_child (state)
			elseif state.child_already_set (Current) then
				-- `is_allowed_child' already called, as I'm the only
				-- remaining child, I find it OK
				Result := True
			else
				-- finish scanning for required childs from the current position
				Result := not any_required_child (state)
			end
		end


feature {CHILD_DEFINITION} -- start tag transition

	is_allowed_set (state: ELEMENT_STATE;
						 next_element: DEFINED_ELEMENT): BOOLEAN is
		do
			Result := child.is_allowed_set (state, next_element)
		end

	is_allowed_stay (state: ELEMENT_STATE;
						  next_element: DEFINED_ELEMENT): BOOLEAN is
		require else
			have_child: state.active_child /= Void
		local
			previous_count: INTEGER
		do
			-- only set for check/debugging purposes
			previous_count := state.children.count

			Result := state.active_child.is_allowed_stay (state, next_element)
			if not Result then
				-- on False, stack should have been decreased by child
				check
					stack_decreases: state.children.count < previous_count
				end
				if not state.authoritative_false then
					if state.child_already_set (Current) then
						-- bottom of stack reached, cleanup
						state.children.remove
					else
						Result := is_allowed_stay (state, next_element)
					end
				end
			end
		end


feature {CHILD_DEFINITION} -- end tag transition

	any_required_child (state: ELEMENT_STATE): BOOLEAN is
			-- return True if, from the position record in `state', there is
			-- a required child definition
		do
			debug
				print ("CHILD_DEFINITION_ROOT.any_required_child.%N")
			end

			Result := state.active_child.any_required_child (state)
			if not Result then
				-- we have to walk the child stack up to the root to check
				-- if there are required elements
				state.children.remove
				if state.child_already_set (Current) then
					-- got my own state, so we're finished
					state.children.remove
				else
					Result := any_required_child (state)
				end
			end
		end


feature {NONE} -- private state

	child: CHILD_DEFINITION


invariant

	has_child: child /= Void


end
