indexing

	description: "Describes a sequence, a comma separated list"


class

	CHILD_DEFINITION_SEQ


inherit

	CHILD_DEFINITION


creation

	make,
	make_from_array


feature {NONE} -- Initialization

	make (a_list: DS_LINKED_LIST [CHILD_DEFINITION]) is
		require
			list_not_void: a_list /= Void
		do
			list := a_list
		end

	make_from_array (a_array: ARRAY [CHILD_DEFINITION]) is
		require
			array_not_void: a_array /= Void
			array_has_at_least_one_element: a_array.count >= 1
		local
			i: INTEGER
		do
			create list.make
			from
				i := a_array.lower
			until
				i > a_array.upper
			loop
				list.put_last (a_array.item (i))
				i:= i + 1
			end
		end


feature -- Public state

	list: DS_LINKED_LIST [CHILD_DEFINITION]
			-- My list of allowed tags.


feature {CHILD_DEFINITION} -- Start tag transition

	is_allowed_set (state: ELEMENT_STATE;
							  next_element: DEFINED_ELEMENT): BOOLEAN is
		local
			child_state: CHILD_STATE
			child: CHILD_DEFINITION
			cursor: DS_LINEAR_CURSOR [CHILD_DEFINITION]
		do
			debug
				print ("CHILD_DEFINITION_SEQ.is_allowed_set%N")
			end

			-- push new state level on stack, this object becomes active
			-- done proactively, so any state state pushed by my children
			-- is top
			create child_state.make (Current)
			state.children.put (child_state)

			cursor := list.new_cursor

			-- as children can be optional, we continue until a set is
			-- allowed or an authoritative false is found
			from
				cursor.start
			until
				Result or else cursor.after or else state.authoritative_false
			loop
				child := cursor.item
				Result := child.is_allowed_set (state, next_element)
				if not Result then
					cursor.forth
				end
			end

			if Result then
				-- we want to stay
				-- we safe cursor to keep track of my current child
				child_state.set_cursor (cursor)
			else
				state.children.remove
			end
		end

	is_allowed_stay (state: ELEMENT_STATE;
						  next_element: DEFINED_ELEMENT): BOOLEAN is
		require
			my_state: state.active_child = Current
		local
			child: CHILD_DEFINITION
			cursor: DS_LINEAR_CURSOR [CHILD_DEFINITION]
		do
			debug
				print ("CHILD_DEFINITION_SEQ.is_allowed_stay%N")
			end

			-- as children can be optional, we continue until a set is
			-- allowed or an authoritative false is found
			from
				cursor := state.current_child_state.cursor
				cursor.forth
			until
				Result or else cursor.after or else state.authoritative_false
			loop
				child := state.current_child_state.cursor.item
				Result := child.is_allowed_set (state, next_element)
				if not Result then
					cursor.forth
				end
			end

			if not Result then
				-- we're no longer active, go to my parent child definition
				state.children.remove
			end
		ensure
			stack_decrease: not Result implies state.children.count < old state.children.count
		end


feature {DEFINED_ELEMENT, CHILD_DEFINITION} -- End tag transition

	any_required_child (state: ELEMENT_STATE): BOOLEAN is
			-- return True if, from the position record in `state', there is
			-- a required child definition
		local
			cursor: DS_LINEAR_CURSOR [CHILD_DEFINITION]
		do
			debug
				print ("CHILD_DEFINITION_SEQ.any_required_child.%N")
			end

			-- if we have been set, check if one of our remaining childs
			-- is required, else we start at the first child
			if state.child_already_set (Current) then

				check
					have_cursor: state.current_child_state.cursor /= Void
				end

				cursor := state.current_child_state.cursor
				-- current item was ok, move to next item
				cursor.forth
			else
				-- is this code reachable?
				cursor := list.new_cursor
				cursor.start
			end

			from
			until
				Result or else cursor.after
			loop
				Result := cursor.item.any_required_child (state)
				cursor.forth
			end
		end


invariant

	list_not_void: list /= Void
	at_least_one_child: not list.is_empty

end
